# Copyright (C) 2017-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, List, Optional, Tuple, Union

from swh.model.hashutil import hash_to_bytes

from .cache import VaultCache

ObjectId = Union[str, bytes]


class InMemoryVaultBackend:
    """Stub vault backend, for use in the CLI."""

    def __init__(self):
        self._cache = VaultCache(cls="memory")

    def fetch(self, obj_type: str, obj_id: ObjectId) -> Optional[bytes]:
        return self._cache.get(obj_type, hash_to_bytes(obj_id))

    def cook(
        self, obj_type: str, obj_id: ObjectId, email: Optional[str] = None
    ) -> Dict[str, Any]:
        raise NotImplementedError("InMemoryVaultBackend.cook()")

    def progress(self, obj_type: str, obj_id: ObjectId):
        raise NotImplementedError("InMemoryVaultBackend.progress()")

    # Cookers endpoints

    def set_progress(self, obj_type: str, obj_id: ObjectId, progress: str) -> None:
        pass

    def set_status(self, obj_type: str, obj_id: ObjectId, status: str) -> None:
        pass

    def put_bundle(self, obj_type: str, obj_id: ObjectId, bundle) -> bool:
        self._cache.add(obj_type, hash_to_bytes(obj_id), bundle)
        return True

    def send_notif(self, obj_type: str, obj_id: ObjectId):
        pass

    # Batch endpoints

    def batch_cook(self, batch: List[Tuple[str, str]]) -> int:
        raise NotImplementedError("InMemoryVaultBackend.batch_cook()")

    def batch_progress(self, batch_id: int) -> Dict[str, Any]:
        pass



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
This cooker creates tarballs containing a bare .git directory,
that can be unpacked and cloned like any git repository.

It works in three steps:

1. Write objects one by one in :file:`.git/objects/`
2. Calls ``git repack`` to pack all these objects into git packfiles.
3. Creates a tarball of the resulting repository

It keeps a set of all written (or about-to-be-written) object hashes in memory
to avoid downloading and writing the same objects twice.
"""

import datetime
import os.path
import re
import subprocess
import tarfile
import tempfile
from typing import Any, Dict, Iterable, List, Set
import zlib

from swh.core.api.classes import stream_results
from swh.graph.client import GraphArgumentException
from swh.model import identifiers
from swh.model.hashutil import hash_to_bytehex, hash_to_hex
from swh.model.model import (
    Person,
    Revision,
    RevisionType,
    Sha1Git,
    TimestampWithTimezone,
)
from swh.storage.algos.revisions_walker import DFSRevisionsWalker
from swh.vault.cookers.base import BaseVaultCooker
from swh.vault.to_disk import HIDDEN_MESSAGE, SKIPPED_MESSAGE

REVISION_BATCH_SIZE = 10000
DIRECTORY_BATCH_SIZE = 10000
CONTENT_BATCH_SIZE = 100


class GitBareCooker(BaseVaultCooker):
    use_fsck = True

    def cache_type_key(self) -> str:
        return self.obj_type

    def check_exists(self):
        obj_type = self.obj_type.split("_")[0]
        if obj_type == "revision":
            return not list(self.storage.revision_missing([self.obj_id]))
        elif obj_type == "directory":
            return not list(self.storage.directory_missing([self.obj_id]))
        else:
            raise NotImplementedError(f"GitBareCooker for {obj_type}")

    def obj_swhid(self) -> identifiers.CoreSWHID:
        obj_type = self.obj_type.split("_")[0]
        return identifiers.CoreSWHID(
            object_type=identifiers.ObjectType[obj_type.upper()], object_id=self.obj_id,
        )

    def _push(self, stack: List[Sha1Git], obj_ids: Iterable[Sha1Git]) -> None:
        assert not isinstance(obj_ids, bytes)
        revision_ids = [id_ for id_ in obj_ids if id_ not in self._seen]
        self._seen.update(revision_ids)
        stack.extend(revision_ids)

    def _pop(self, stack: List[Sha1Git], n: int) -> List[Sha1Git]:
        obj_ids = stack[-n:]
        stack[-n:] = []
        return obj_ids

    def prepare_bundle(self):
        # Objects we will visit soon:
        self._rev_stack: List[Sha1Git] = []
        self._dir_stack: List[Sha1Git] = []
        self._cnt_stack: List[Sha1Git] = []

        # Set of objects already in any of the stacks:
        self._seen: Set[Sha1Git] = set()

        # Set of errors we expect git-fsck to raise at the end:
        self._expected_fsck_errors = set()

        with tempfile.TemporaryDirectory(prefix="swh-vault-gitbare-") as workdir:
            # Initialize a Git directory
            self.workdir = workdir
            self.gitdir = os.path.join(workdir, "clone.git")
            os.mkdir(self.gitdir)
            self.init_git()

            # Add the root object to the stack of objects to visit
            self.push_subgraph(self.obj_type.split("_")[0], self.obj_id)

            # Load and write all the objects to disk
            self.load_objects()

            # Write the root object as a ref.
            # This must be done before repacking; git-repack ignores orphan objects.
            self.write_refs()

            self.repack()
            self.write_archive()

    def init_git(self) -> None:
        subprocess.run(["git", "-C", self.gitdir, "init", "--bare"], check=True)

        # Create all possible dirs ahead of time, so we don't have to check for
        # existence every time.
        for byte in range(256):
            os.mkdir(os.path.join(self.gitdir, "objects", f"{byte:02x}"))

    def repack(self) -> None:
        if self.use_fsck:
            self.git_fsck()

        # Add objects we wrote in a pack
        subprocess.run(["git", "-C", self.gitdir, "repack"], check=True)

        # Remove their non-packed originals
        subprocess.run(["git", "-C", self.gitdir, "prune-packed"], check=True)

    def git_fsck(self) -> None:
        proc = subprocess.run(
            ["git", "-C", self.gitdir, "fsck"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env={"LANG": "C.utf8"},
        )
        if not self._expected_fsck_errors:
            # All went well, there should not be any error
            proc.check_returncode()
            return

        # Split on newlines not followed by a space
        errors = re.split("\n(?! )", proc.stdout.decode())

        unexpected_errors = set(filter(bool, errors)) - self._expected_fsck_errors
        if unexpected_errors:
            raise Exception(
                "\n".join(
                    ["Unexpected errors from git-fsck:"] + sorted(unexpected_errors)
                )
            )

    def write_refs(self):
        obj_type = self.obj_type.split("_")[0]
        if obj_type == "directory":
            # We need a synthetic revision pointing to the directory
            author = Person.from_fullname(
                b"swh-vault, git-bare cooker <robot@softwareheritage.org>"
            )
            dt = datetime.datetime.now(tz=datetime.timezone.utc)
            dt = dt.replace(microsecond=0)  # not supported by git
            date = TimestampWithTimezone.from_datetime(dt)
            revision = Revision(
                author=author,
                committer=author,
                date=date,
                committer_date=date,
                message=b"Initial commit",
                type=RevisionType.GIT,
                directory=self.obj_id,
                synthetic=True,
            )
            self.write_revision_node(revision.to_dict())
            head = revision.id
        elif obj_type == "revision":
            head = self.obj_id
        else:
            assert False, obj_type

        with open(os.path.join(self.gitdir, "refs", "heads", "master"), "wb") as fd:
            fd.write(hash_to_bytehex(head))

    def write_archive(self):
        with tarfile.TarFile(mode="w", fileobj=self.fileobj) as tf:
            tf.add(self.gitdir, arcname=f"{self.obj_swhid()}.git", recursive=True)

    def _obj_path(self, obj_id: Sha1Git):
        return os.path.join(self.gitdir, self._obj_relative_path(obj_id))

    def _obj_relative_path(self, obj_id: Sha1Git):
        obj_id_hex = hash_to_hex(obj_id)
        directory = obj_id_hex[0:2]
        filename = obj_id_hex[2:]
        return os.path.join("objects", directory, filename)

    def object_exists(self, obj_id: Sha1Git) -> bool:
        return os.path.exists(self._obj_path(obj_id))

    def write_object(self, obj_id: Sha1Git, obj: bytes) -> bool:
        """Writes a git object on disk.

        Returns whether it was already written."""
        # Git requires objects to be zlib-compressed; but repacking decompresses and
        # removes them, so we don't need to compress them too much.
        data = zlib.compress(obj, level=1)

        with open(self._obj_path(obj_id), "wb") as fd:
            fd.write(data)
        return True

    def push_subgraph(self, obj_type, obj_id) -> None:
        if obj_type == "revision":
            self.push_revision_subgraph(obj_id)
        elif obj_type == "directory":
            self._push(self._dir_stack, [obj_id])
        else:
            raise NotImplementedError(
                f"GitBareCooker.queue_subgraph({obj_type!r}, ...)"
            )

    def load_objects(self) -> None:
        while self._rev_stack or self._dir_stack or self._cnt_stack:
            revision_ids = self._pop(self._rev_stack, REVISION_BATCH_SIZE)
            self.load_revisions(revision_ids)

            directory_ids = self._pop(self._dir_stack, DIRECTORY_BATCH_SIZE)
            self.load_directories(directory_ids)

            content_ids = self._pop(self._cnt_stack, CONTENT_BATCH_SIZE)
            self.load_contents(content_ids)

    def push_revision_subgraph(self, obj_id: Sha1Git) -> None:
        """Fetches a revision and all its children, and writes them to disk"""
        loaded_from_graph = False

        if self.graph:
            # First, try to cook using swh-graph, as it is more efficient than
            # swh-storage for querying the history
            obj_swhid = identifiers.CoreSWHID(
                object_type=identifiers.ObjectType.REVISION, object_id=obj_id,
            )
            try:
                revision_ids = (
                    swhid.object_id
                    for swhid in map(
                        identifiers.CoreSWHID.from_string,
                        self.graph.visit_nodes(str(obj_swhid), edges="rev:rev"),
                    )
                )
                self._push(self._rev_stack, revision_ids)
            except GraphArgumentException:
                # Revision not found in the graph
                pass
            else:
                loaded_from_graph = True

        if not loaded_from_graph:
            # If swh-graph is not available, or the revision is not yet in
            # swh-graph, fall back to self.storage.revision_log.
            # self.storage.revision_log also gives us the full revisions,
            # so we load them right now instead of just pushing them on the stack.
            walker = DFSRevisionsWalker(self.storage, obj_id)
            for revision in walker:
                self.write_revision_node(revision)
                self._push(self._dir_stack, [revision["directory"]])

    def load_revisions(self, obj_ids: List[Sha1Git]) -> None:
        """Given a list of revision ids, loads these revisions and their directories;
        but not their parent revisions."""
        revisions = self.storage.revision_get(obj_ids)
        for revision in revisions:
            self.write_revision_node(revision.to_dict())
        self._push(self._dir_stack, (rev.directory for rev in revisions))

    def write_revision_node(self, revision: Dict[str, Any]) -> bool:
        """Writes a revision object to disk"""
        git_object = identifiers.revision_git_object(revision)
        return self.write_object(revision["id"], git_object)

    def load_directories(self, obj_ids: List[Sha1Git]) -> None:
        for obj_id in obj_ids:
            self.load_directory(obj_id)

    def load_directory(self, obj_id: Sha1Git) -> None:
        # Load the directory
        entries = [
            entry.to_dict()
            for entry in stream_results(self.storage.directory_get_entries, obj_id)
        ]
        directory = {"id": obj_id, "entries": entries}
        git_object = identifiers.directory_git_object(directory)
        self.write_object(obj_id, git_object)

        # Add children to the stack
        entry_loaders: Dict[str, List[Sha1Git]] = {
            "file": self._cnt_stack,
            "dir": self._dir_stack,
            "rev": self._rev_stack,
        }
        for entry in directory["entries"]:
            stack = entry_loaders[entry["type"]]
            self._push(stack, [entry["target"]])

    def load_contents(self, obj_ids: List[Sha1Git]) -> None:
        # TODO: add support of filtered objects, somehow?
        # It's tricky, because, by definition, we can't write a git object with
        # the expected hash, so git-fsck *will* choke on it.
        contents = self.storage.content_get(obj_ids, "sha1_git")

        visible_contents = []
        for (obj_id, content) in zip(obj_ids, contents):
            if content is None:
                # FIXME: this may also happen for missing content
                self.write_content(obj_id, SKIPPED_MESSAGE)
                self._expect_mismatched_object_error(obj_id)
            elif content.status == "visible":
                visible_contents.append(content)
            elif content.status == "hidden":
                self.write_content(obj_id, HIDDEN_MESSAGE)
                self._expect_mismatched_object_error(obj_id)
            else:
                assert False, (
                    f"unexpected status {content.status!r} "
                    f"for content {hash_to_hex(content.sha1_git)}"
                )

        if self.objstorage is None:
            for content in visible_contents:
                data = self.storage.content_get_data(content.sha1)
                self.write_content(content.sha1_git, data)
        else:
            content_data = self.objstorage.get_batch(c.sha1 for c in visible_contents)
            for (content, data) in zip(contents, content_data):
                self.write_content(content.sha1_git, data)

    def write_content(self, obj_id: Sha1Git, content: bytes) -> None:
        header = identifiers.git_object_header("blob", len(content))
        self.write_object(obj_id, header + content)

    def _expect_mismatched_object_error(self, obj_id):
        obj_id_hex = hash_to_hex(obj_id)
        obj_path = self._obj_relative_path(obj_id)
        self._expected_fsck_errors.add(
            f"error: sha1 mismatch for ./{obj_path} (expected {obj_id_hex})"
        )
        self._expected_fsck_errors.add(
            f"error: {obj_id_hex}: object corrupt or missing: ./{obj_path}"
        )
        self._expected_fsck_errors.add(f"missing blob {obj_id_hex}")



# Copyright (C) 2015-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import logging
from typing import TYPE_CHECKING, Optional

import click

from swh.core.cli import CONTEXT_SETTINGS, AliasedGroup
from swh.core.cli import swh as swh_cli_group

if TYPE_CHECKING:
    import io

    from swh.model.identifiers import CoreSWHID


class SwhidParamType(click.ParamType):
    name = "swhid"

    def convert(self, value, param, ctx):
        from swh.model.exceptions import ValidationError
        from swh.model.identifiers import CoreSWHID

        try:
            return CoreSWHID.from_string(value)
        except ValidationError:
            self.fail(f"expected core SWHID, got {value!r}", param, ctx)


@swh_cli_group.group(name="vault", context_settings=CONTEXT_SETTINGS, cls=AliasedGroup)
@click.pass_context
def vault(ctx):
    """Software Heritage Vault tools."""


@vault.command()
@click.option(
    "--config-file",
    "-C",
    default=None,
    metavar="CONFIGFILE",
    type=click.Path(exists=True, dir_okay=False,),
    help="Configuration file.",
)
@click.argument("swhid", type=SwhidParamType())
@click.argument("outfile", type=click.File("wb"))
@click.option(
    "--cooker-type",
    type=click.Choice(["flat", "gitfast", "git_bare"]),
    help="Selects which cooker to use, when there is more than one available "
    "for the given object type.",
)
@click.pass_context
def cook(
    ctx,
    config_file: str,
    swhid: CoreSWHID,
    outfile: io.RawIOBase,
    cooker_type: Optional[str],
):
    """
    Runs a vault cooker for a single object (identified by a SWHID),
    and outputs it to the given file.
    """
    from swh.core import config
    from swh.objstorage.factory import get_objstorage
    from swh.storage import get_storage

    from .cookers import COOKER_TYPES, get_cooker_cls
    from .in_memory_backend import InMemoryVaultBackend

    conf = config.read(config_file)

    supported_object_types = {name.split("_")[0] for name in COOKER_TYPES}
    if swhid.object_type.name.lower() not in supported_object_types:
        raise click.ClickException(
            f"No cooker available for {swhid.object_type.name} objects."
        )

    cooker_name = swhid.object_type.name.lower()

    if cooker_type:
        cooker_name = f"{cooker_name}_{cooker_type}"
        if cooker_name not in COOKER_TYPES:
            raise click.ClickException(
                f"{swhid.object_type.name.lower()} objects do not have "
                f"a {cooker_type} cooker."
            )
    else:
        if cooker_name not in COOKER_TYPES:
            raise click.ClickException(
                f"{swhid.object_type.name.lower()} objects need "
                f"an explicit --cooker-type."
            )

    try:
        from swh.graph.client import RemoteGraphClient  # optional dependency

        graph = RemoteGraphClient(**conf["graph"]) if conf.get("graph") else None
    except ModuleNotFoundError:
        if conf.get("graph"):
            raise EnvironmentError(
                "Graph configuration required but module is not installed."
            )
        else:
            graph = None

    backend = InMemoryVaultBackend()
    storage = get_storage(**conf["storage"])
    objstorage = get_objstorage(**conf["objstorage"]) if "objstorage" in conf else None
    cooker_cls = get_cooker_cls(cooker_name)
    cooker = cooker_cls(
        obj_type=cooker_name,
        obj_id=swhid.object_id,
        backend=backend,
        storage=storage,
        graph=graph,
        objstorage=objstorage,
    )
    cooker.cook()

    bundle = backend.fetch(cooker_name, swhid.object_id)
    assert bundle, "Cooker did not write a bundle to the backend."
    outfile.write(bundle)


@vault.command(name="rpc-serve")
@click.option(
    "--config-file",
    "-C",
    default=None,
    metavar="CONFIGFILE",
    type=click.Path(exists=True, dir_okay=False,),
    help="Configuration file.",
)
@click.option(
    "--host",
    default="0.0.0.0",
    metavar="IP",
    show_default=True,
    help="Host ip address to bind the server on",
)
@click.option(
    "--port",
    default=5005,
    type=click.INT,
    metavar="PORT",
    help="Binding port of the server",
)
@click.option(
    "--debug/--no-debug",
    default=True,
    help="Indicates if the server should run in debug mode",
)
@click.pass_context
def serve(ctx, config_file, host, port, debug):
    """Software Heritage Vault RPC server."""
    import aiohttp

    from swh.vault.api.server import make_app_from_configfile

    ctx.ensure_object(dict)

    try:
        app = make_app_from_configfile(config_file, debug=debug)
    except EnvironmentError as e:
        click.echo(e.msg, err=True)
        ctx.exit(1)

    aiohttp.web.run_app(app, host=host, port=int(port))


def main():
    logging.basicConfig()
    return serve(auto_envvar_prefix="SWH_VAULT")


if __name__ == "__main__":
    main()



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import contextlib
import datetime
from unittest.mock import MagicMock, patch

import attr
import psycopg2
import pytest

from swh.model import hashutil
from swh.vault.exc import NotFoundExc
from swh.vault.tests.vault_testing import hash_content


@contextlib.contextmanager
def mock_cooking(vault_backend):
    with patch.object(vault_backend, "_send_task") as mt:
        mt.return_value = 42
        with patch("swh.vault.backend.get_cooker_cls") as mg:
            mcc = MagicMock()
            mc = MagicMock()
            mg.return_value = mcc
            mcc.return_value = mc
            mc.check_exists.return_value = True

            yield {
                "_send_task": mt,
                "get_cooker_cls": mg,
                "cooker_cls": mcc,
                "cooker": mc,
            }


def assertTimestampAlmostNow(ts, tolerance_secs=1.0):  # noqa
    now = datetime.datetime.now(datetime.timezone.utc)
    creation_delta_secs = (ts - now).total_seconds()
    assert creation_delta_secs < tolerance_secs


def fake_cook(backend, obj_type, result_content, sticky=False):
    content, obj_id = hash_content(result_content)
    with mock_cooking(backend):
        backend.create_task(obj_type, obj_id, sticky)
    backend.cache.add(obj_type, obj_id, b"content")
    backend.set_status(obj_type, obj_id, "done")
    return obj_id, content


def fail_cook(backend, obj_type, obj_id, failure_reason):
    with mock_cooking(backend):
        backend.create_task(obj_type, obj_id)
    backend.set_status(obj_type, obj_id, "failed")
    backend.set_progress(obj_type, obj_id, failure_reason)


TEST_TYPE = "revision_gitfast"
TEST_HEX_ID = "4a4b9771542143cf070386f86b4b92d42966bdbc"
TEST_OBJ_ID = hashutil.hash_to_bytes(TEST_HEX_ID)
TEST_PROGRESS = (
    "Mr. White, You're telling me you're cooking again?" " \N{ASTONISHED FACE} "
)
TEST_EMAIL = "ouiche@lorraine.fr"


@pytest.fixture
def swh_vault(swh_vault, sample_data):
    # make the vault's storage consistent with test data
    revision = attr.evolve(sample_data.revision, id=TEST_OBJ_ID)
    swh_vault.storage.revision_add([revision])
    return swh_vault


def test_create_task_simple(swh_vault):
    with mock_cooking(swh_vault) as m:
        swh_vault.create_task(TEST_TYPE, TEST_HEX_ID)

    m["get_cooker_cls"].assert_called_once_with(TEST_TYPE)

    args = m["cooker_cls"].call_args[0]
    assert args[0] == TEST_TYPE
    assert args[1] == TEST_HEX_ID

    assert m["cooker"].check_exists.call_count == 1
    assert m["_send_task"].call_count == 1

    args = m["_send_task"].call_args[0]
    assert args[0] == TEST_TYPE
    assert args[1] == TEST_HEX_ID

    info = swh_vault.progress(TEST_TYPE, TEST_HEX_ID)
    assert info["object_id"] == TEST_HEX_ID
    assert info["type"] == TEST_TYPE
    assert info["task_status"] == "new"
    assert info["task_id"] == 42

    assertTimestampAlmostNow(info["ts_created"])

    assert info["ts_done"] is None
    assert info["progress_msg"] is None


def test_create_fail_duplicate_task(swh_vault):
    with mock_cooking(swh_vault):
        swh_vault.create_task(TEST_TYPE, TEST_HEX_ID)
        with pytest.raises(psycopg2.IntegrityError):
            swh_vault.create_task(TEST_TYPE, TEST_HEX_ID)


def test_create_fail_nonexisting_object(swh_vault):
    with mock_cooking(swh_vault) as m:
        m["cooker"].check_exists.side_effect = ValueError("Nothing here.")
        with pytest.raises(ValueError):
            swh_vault.create_task(TEST_TYPE, TEST_HEX_ID)


def test_create_set_progress(swh_vault):
    with mock_cooking(swh_vault):
        swh_vault.create_task(TEST_TYPE, TEST_HEX_ID)

    info = swh_vault.progress(TEST_TYPE, TEST_HEX_ID)
    assert info["progress_msg"] is None
    swh_vault.set_progress(TEST_TYPE, TEST_HEX_ID, TEST_PROGRESS)
    info = swh_vault.progress(TEST_TYPE, TEST_HEX_ID)
    assert info["progress_msg"] == TEST_PROGRESS


def test_create_set_status(swh_vault):
    with mock_cooking(swh_vault):
        swh_vault.create_task(TEST_TYPE, TEST_HEX_ID)

    info = swh_vault.progress(TEST_TYPE, TEST_HEX_ID)
    assert info["task_status"] == "new"
    assert info["ts_done"] is None

    swh_vault.set_status(TEST_TYPE, TEST_HEX_ID, "pending")
    info = swh_vault.progress(TEST_TYPE, TEST_HEX_ID)
    assert info["task_status"] == "pending"
    assert info["ts_done"] is None

    swh_vault.set_status(TEST_TYPE, TEST_HEX_ID, "done")
    info = swh_vault.progress(TEST_TYPE, TEST_HEX_ID)
    assert info["task_status"] == "done"
    assertTimestampAlmostNow(info["ts_done"])


def test_create_update_access_ts(swh_vault):
    with mock_cooking(swh_vault):
        swh_vault.create_task(TEST_TYPE, TEST_OBJ_ID)

    info = swh_vault.progress(TEST_TYPE, TEST_OBJ_ID)
    access_ts_1 = info["ts_last_access"]
    assertTimestampAlmostNow(access_ts_1)

    swh_vault.update_access_ts(TEST_TYPE, TEST_OBJ_ID)
    info = swh_vault.progress(TEST_TYPE, TEST_OBJ_ID)
    access_ts_2 = info["ts_last_access"]
    assertTimestampAlmostNow(access_ts_2)

    swh_vault.update_access_ts(TEST_TYPE, TEST_OBJ_ID)
    info = swh_vault.progress(TEST_TYPE, TEST_OBJ_ID)

    access_ts_3 = info["ts_last_access"]
    assertTimestampAlmostNow(access_ts_3)

    assert access_ts_1 < access_ts_2
    assert access_ts_2 < access_ts_3


def test_cook_idempotent(swh_vault, sample_data):
    with mock_cooking(swh_vault):
        info1 = swh_vault.cook(TEST_TYPE, TEST_HEX_ID)
        info2 = swh_vault.cook(TEST_TYPE, TEST_HEX_ID)
        info3 = swh_vault.cook(TEST_TYPE, TEST_HEX_ID)
        assert info1 == info2
        assert info1 == info3


def test_cook_email_pending_done(swh_vault):
    with mock_cooking(swh_vault), patch.object(
        swh_vault, "add_notif_email"
    ) as madd, patch.object(swh_vault, "send_notification") as msend:

        swh_vault.cook(TEST_TYPE, TEST_HEX_ID)
        madd.assert_not_called()
        msend.assert_not_called()

        madd.reset_mock()
        msend.reset_mock()

        swh_vault.cook(TEST_TYPE, TEST_HEX_ID, email=TEST_EMAIL)
        madd.assert_called_once_with(TEST_TYPE, TEST_OBJ_ID, TEST_EMAIL)
        msend.assert_not_called()

        madd.reset_mock()
        msend.reset_mock()

        swh_vault.set_status(TEST_TYPE, TEST_HEX_ID, "done")
        swh_vault.cook(TEST_TYPE, TEST_HEX_ID, email=TEST_EMAIL)
        msend.assert_called_once_with(None, TEST_EMAIL, TEST_TYPE, TEST_HEX_ID, "done")
        madd.assert_not_called()


def test_send_all_emails(swh_vault):
    with mock_cooking(swh_vault):
        emails = ("a@example.com", "billg@example.com", "test+42@example.org")
        for email in emails:
            swh_vault.cook(TEST_TYPE, TEST_HEX_ID, email=email)

    swh_vault.set_status(TEST_TYPE, TEST_HEX_ID, "done")

    with patch.object(swh_vault, "smtp_server") as m:
        swh_vault.send_notif(TEST_TYPE, TEST_HEX_ID)

        sent_emails = {k[0][0] for k in m.send_message.call_args_list}
        assert {k["To"] for k in sent_emails} == set(emails)

        for e in sent_emails:
            assert "bot@softwareheritage.org" in e["From"]
            assert TEST_TYPE in e["Subject"]
            assert TEST_HEX_ID[:5] in e["Subject"]
            assert TEST_TYPE in str(e)
            assert "https://archive.softwareheritage.org/" in str(e)
            assert TEST_HEX_ID[:5] in str(e)
            assert "--\x20\n" in str(e)  # Well-formated signature!!!

        # Check that the entries have been deleted and recalling the
        # function does not re-send the e-mails
        m.reset_mock()
        swh_vault.send_notif(TEST_TYPE, TEST_HEX_ID)
        m.assert_not_called()


def test_available(swh_vault):
    assert not swh_vault.is_available(TEST_TYPE, TEST_HEX_ID)

    with mock_cooking(swh_vault):
        swh_vault.create_task(TEST_TYPE, TEST_HEX_ID)
    assert not swh_vault.is_available(TEST_TYPE, TEST_HEX_ID)

    swh_vault.cache.add(TEST_TYPE, TEST_HEX_ID, b"content")
    assert not swh_vault.is_available(TEST_TYPE, TEST_HEX_ID)

    swh_vault.set_status(TEST_TYPE, TEST_HEX_ID, "done")
    assert swh_vault.is_available(TEST_TYPE, TEST_HEX_ID)


def test_fetch(swh_vault):
    assert swh_vault.fetch(TEST_TYPE, TEST_HEX_ID, raise_notfound=False) is None

    with pytest.raises(
        NotFoundExc, match=f"{TEST_TYPE} {TEST_HEX_ID} is not available."
    ):
        swh_vault.fetch(TEST_TYPE, TEST_HEX_ID)

    obj_id, content = fake_cook(swh_vault, TEST_TYPE, b"content")

    info = swh_vault.progress(TEST_TYPE, obj_id)
    access_ts_before = info["ts_last_access"]

    assert swh_vault.fetch(TEST_TYPE, obj_id) == b"content"

    info = swh_vault.progress(TEST_TYPE, obj_id)
    access_ts_after = info["ts_last_access"]

    assertTimestampAlmostNow(access_ts_after)
    assert access_ts_before < access_ts_after


def test_cache_expire_oldest(swh_vault):
    r = range(1, 10)
    inserted = {}
    for i in r:
        sticky = i == 5
        content = b"content%s" % str(i).encode()
        obj_id, content = fake_cook(swh_vault, TEST_TYPE, content, sticky)
        inserted[i] = (obj_id, content)

    swh_vault.update_access_ts(TEST_TYPE, inserted[2][0])
    swh_vault.update_access_ts(TEST_TYPE, inserted[3][0])
    swh_vault.cache_expire_oldest(n=4)

    should_be_still_here = {2, 3, 5, 8, 9}
    for i in r:
        assert swh_vault.is_available(TEST_TYPE, inserted[i][0]) == (
            i in should_be_still_here
        )


def test_cache_expire_until(swh_vault):
    r = range(1, 10)
    inserted = {}
    for i in r:
        sticky = i == 5
        content = b"content%s" % str(i).encode()
        obj_id, content = fake_cook(swh_vault, TEST_TYPE, content, sticky)
        inserted[i] = (obj_id, content)

        if i == 7:
            cutoff_date = datetime.datetime.now()

    swh_vault.update_access_ts(TEST_TYPE, inserted[2][0])
    swh_vault.update_access_ts(TEST_TYPE, inserted[3][0])
    swh_vault.cache_expire_until(date=cutoff_date)

    should_be_still_here = {2, 3, 5, 8, 9}
    for i in r:
        assert swh_vault.is_available(TEST_TYPE, inserted[i][0]) == (
            i in should_be_still_here
        )


def test_fail_cook_simple(swh_vault):
    fail_cook(swh_vault, TEST_TYPE, TEST_HEX_ID, "error42")
    assert not swh_vault.is_available(TEST_TYPE, TEST_HEX_ID)
    info = swh_vault.progress(TEST_TYPE, TEST_HEX_ID)
    assert info["progress_msg"] == "error42"


def test_send_failure_email(swh_vault):
    with mock_cooking(swh_vault):
        swh_vault.cook(TEST_TYPE, TEST_HEX_ID, email="a@example.com")

    swh_vault.set_status(TEST_TYPE, TEST_HEX_ID, "failed")
    swh_vault.set_progress(TEST_TYPE, TEST_HEX_ID, "test error")

    with patch.object(swh_vault, "smtp_server") as m:
        swh_vault.send_notif(TEST_TYPE, TEST_HEX_ID)

        e = [k[0][0] for k in m.send_message.call_args_list][0]
        assert e["To"] == "a@example.com"

        assert "bot@softwareheritage.org" in e["From"]
        assert TEST_TYPE in e["Subject"]
        assert TEST_HEX_ID[:5] in e["Subject"]
        assert "fail" in e["Subject"]
        assert TEST_TYPE in str(e)
        assert TEST_HEX_ID[:5] in str(e)
        assert "test error" in str(e)
        assert "--\x20\n" in str(e)  # Well-formated signature


def test_retry_failed_bundle(swh_vault):
    fail_cook(swh_vault, TEST_TYPE, TEST_HEX_ID, "error42")
    info = swh_vault.progress(TEST_TYPE, TEST_HEX_ID)
    assert info["task_status"] == "failed"
    with mock_cooking(swh_vault):
        swh_vault.cook(TEST_TYPE, TEST_HEX_ID)
    info = swh_vault.progress(TEST_TYPE, TEST_HEX_ID)
    assert info["task_status"] == "new"



# Copyright (C) 2016-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.model import hashutil
from swh.objstorage.factory import get_objstorage
from swh.objstorage.objstorage import compute_hash


class VaultCache:
    """The Vault cache is an object storage that stores Vault bundles.

    This implementation computes sha1('<bundle_type>:<object_id>') as the
    internal identifiers used in the underlying objstorage.
    """

    def __init__(self, **objstorage):
        self.objstorage = get_objstorage(**objstorage)

    def add(self, obj_type, obj_id, content):
        sid = self._get_internal_id(obj_type, obj_id)
        return self.objstorage.add(content, sid)

    def get(self, obj_type, obj_id):
        sid = self._get_internal_id(obj_type, obj_id)
        return self.objstorage.get(hashutil.hash_to_bytes(sid))

    def delete(self, obj_type, obj_id):
        sid = self._get_internal_id(obj_type, obj_id)
        return self.objstorage.delete(hashutil.hash_to_bytes(sid))

    def add_stream(self, obj_type, obj_id, content_iter):
        sid = self._get_internal_id(obj_type, obj_id)
        return self.objstorage.add_stream(content_iter, sid)

    def get_stream(self, obj_type, obj_id):
        sid = self._get_internal_id(obj_type, obj_id)
        return self.objstorage.get_stream(hashutil.hash_to_bytes(sid))

    def is_cached(self, obj_type, obj_id):
        sid = self._get_internal_id(obj_type, obj_id)
        return hashutil.hash_to_bytes(sid) in self.objstorage

    def _get_internal_id(self, obj_type, obj_id):
        obj_id = hashutil.hash_to_hex(obj_id)
        return compute_hash("{}:{}".format(obj_type, obj_id).encode())



# Copyright (C) 2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.model import hashutil

TEST_TYPE_1 = "revision_gitfast"
TEST_TYPE_2 = "directory"

TEST_HEX_ID_1 = "4a4b9771542143cf070386f86b4b92d42966bdbc"
TEST_HEX_ID_2 = "17a3e48bce37be5226490e750202ad3a9a1a3fe9"

TEST_OBJ_ID_1 = hashutil.hash_to_bytes(TEST_HEX_ID_1)
TEST_OBJ_ID_2 = hashutil.hash_to_bytes(TEST_HEX_ID_2)

TEST_CONTENT_1 = b"test content 1"
TEST_CONTENT_2 = b"test content 2"


# Let's try to avoid replicating edge-cases already tested in
# swh-objstorage, and instead focus on testing behaviors specific to the
# Vault cache here.


def test_internal_id(swh_vault):
    sid = swh_vault.cache._get_internal_id(TEST_TYPE_1, TEST_OBJ_ID_1)
    assert hashutil.hash_to_hex(sid) == "6829cda55b54c295aa043a611a4e0320239988d9"


def test_simple_add_get(swh_vault):
    swh_vault.cache.add(TEST_TYPE_1, TEST_OBJ_ID_1, TEST_CONTENT_1)
    assert swh_vault.cache.get(TEST_TYPE_1, TEST_OBJ_ID_1) == TEST_CONTENT_1
    assert swh_vault.cache.is_cached(TEST_TYPE_1, TEST_OBJ_ID_1)


def test_different_type_same_id(swh_vault):
    swh_vault.cache.add(TEST_TYPE_1, TEST_OBJ_ID_1, TEST_CONTENT_1)
    swh_vault.cache.add(TEST_TYPE_2, TEST_OBJ_ID_1, TEST_CONTENT_2)
    assert swh_vault.cache.get(TEST_TYPE_1, TEST_OBJ_ID_1) == TEST_CONTENT_1
    assert swh_vault.cache.get(TEST_TYPE_2, TEST_OBJ_ID_1) == TEST_CONTENT_2
    assert swh_vault.cache.is_cached(TEST_TYPE_1, TEST_OBJ_ID_1)
    assert swh_vault.cache.is_cached(TEST_TYPE_2, TEST_OBJ_ID_1)


def test_different_type_same_content(swh_vault):
    swh_vault.cache.add(TEST_TYPE_1, TEST_OBJ_ID_1, TEST_CONTENT_1)
    swh_vault.cache.add(TEST_TYPE_2, TEST_OBJ_ID_1, TEST_CONTENT_1)
    assert swh_vault.cache.get(TEST_TYPE_1, TEST_OBJ_ID_1) == TEST_CONTENT_1
    assert swh_vault.cache.get(TEST_TYPE_2, TEST_OBJ_ID_1) == TEST_CONTENT_1
    assert swh_vault.cache.is_cached(TEST_TYPE_1, TEST_OBJ_ID_1)
    assert swh_vault.cache.is_cached(TEST_TYPE_2, TEST_OBJ_ID_1)


def test_different_id_same_type(swh_vault):
    swh_vault.cache.add(TEST_TYPE_1, TEST_OBJ_ID_1, TEST_CONTENT_1)
    swh_vault.cache.add(TEST_TYPE_1, TEST_OBJ_ID_2, TEST_CONTENT_2)
    assert swh_vault.cache.get(TEST_TYPE_1, TEST_OBJ_ID_1) == TEST_CONTENT_1
    assert swh_vault.cache.get(TEST_TYPE_1, TEST_OBJ_ID_2) == TEST_CONTENT_2
    assert swh_vault.cache.is_cached(TEST_TYPE_1, TEST_OBJ_ID_1)
    assert swh_vault.cache.is_cached(TEST_TYPE_1, TEST_OBJ_ID_2)



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

import importlib
import logging
from typing import Dict
import warnings

logger = logging.getLogger(__name__)


BACKEND_TYPES: Dict[str, str] = {
    "remote": ".api.client.RemoteVaultClient",
    "local": ".backend.VaultBackend",
    "memory": ".in_memory_backend.InMemoryVaultBackend",
}


def get_vault(cls: str = "remote", **kwargs):
    """
    Get a vault object of class `vault_class` with arguments
    `vault_args`.

    Args:
        cls: vault's class, either 'remote' or 'local'
        kwargs: arguments to pass to the class' constructor

    Returns:
        an instance of VaultBackend (either local or remote)

    Raises:
        ValueError if passed an unknown storage class.

    """
    if "args" in kwargs:
        warnings.warn(
            'Explicit "args" key is deprecated, use keys directly instead.',
            DeprecationWarning,
        )
        kwargs = kwargs["args"]

    class_path = BACKEND_TYPES.get(cls)
    if class_path is None:
        raise ValueError(
            f"Unknown Vault class `{cls}`. " f"Supported: {', '.join(BACKEND_TYPES)}"
        )

    (module_path, class_name) = class_path.rsplit(".", 1)
    module = importlib.import_module(module_path, package=__package__)
    Vault = getattr(module, class_name)
    return Vault(**kwargs)



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
This module contains additional tests for the bare cooker.
Generic cooker tests (eg. without swh-graph) in test_cookers.py also
run on the bare cooker.
"""

import datetime
import io
import subprocess
import tarfile
import tempfile
import unittest.mock

import pytest

from swh.graph.naive_client import NaiveClient as GraphClient
from swh.model.from_disk import DentryPerms
from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    Person,
    Revision,
    RevisionType,
    TimestampWithTimezone,
)
from swh.vault.cookers.git_bare import GitBareCooker
from swh.vault.in_memory_backend import InMemoryVaultBackend


def get_objects(last_revision_in_graph):
    """
    Build objects::

          rev1  <------ rev2
           |             |
           v             v
          dir1          dir2
           |           /   |
           v          /    v
          cnt1  <----°    cnt2
    """
    date = TimestampWithTimezone.from_datetime(
        datetime.datetime(2021, 5, 7, 8, 43, 59, tzinfo=datetime.timezone.utc)
    )
    author = Person.from_fullname(b"Foo <foo@example.org>")
    cnt1 = Content.from_data(b"hello")
    cnt2 = Content.from_data(b"world")
    dir1 = Directory(
        entries=(
            DirectoryEntry(
                name=b"file1",
                type="file",
                perms=DentryPerms.content,
                target=cnt1.sha1_git,
            ),
        )
    )
    dir2 = Directory(
        entries=(
            DirectoryEntry(
                name=b"file1",
                type="file",
                perms=DentryPerms.content,
                target=cnt1.sha1_git,
            ),
            DirectoryEntry(
                name=b"file2",
                type="file",
                perms=DentryPerms.content,
                target=cnt2.sha1_git,
            ),
        )
    )
    rev1 = Revision(
        message=b"msg1",
        date=date,
        committer_date=date,
        author=author,
        committer=author,
        directory=dir1.id,
        type=RevisionType.GIT,
        synthetic=True,
    )
    rev2 = Revision(
        message=b"msg2",
        date=date,
        committer_date=date,
        author=author,
        committer=author,
        directory=dir2.id,
        parents=(rev1.id,),
        type=RevisionType.GIT,
        synthetic=True,
    )

    if last_revision_in_graph:
        nodes = [str(n.swhid()) for n in [cnt1, cnt2, dir1, dir2, rev1, rev2]]
        edges = [
            (str(s.swhid()), str(d.swhid()))
            for (s, d) in [
                (dir1, cnt1),
                (dir2, cnt1),
                (dir2, cnt2),
                (rev1, dir1),
                (rev2, dir2),
                (rev2, rev1),
            ]
        ]
    else:
        nodes = [str(n.swhid()) for n in [cnt1, cnt2, dir1, dir2, rev1]]
        edges = [
            (str(s.swhid()), str(d.swhid()))
            for (s, d) in [(dir1, cnt1), (dir2, cnt1), (dir2, cnt2), (rev1, dir1),]
        ]

    return (cnt1, cnt2, dir1, dir2, rev1, rev2, nodes, edges)


@pytest.mark.parametrize("last_revision_in_graph", [True, False])
def test_graph_revisions(swh_storage, last_revision_in_graph):
    (cnt1, cnt2, dir1, dir2, rev1, rev2, nodes, edges) = get_objects(
        last_revision_in_graph
    )

    # Add all objects to storage
    swh_storage.content_add([cnt1, cnt2])
    swh_storage.directory_add([dir1, dir2])
    swh_storage.revision_add([rev1, rev2])

    # Add spy on swh_storage, to make sure revision_log is not called
    # (the graph must be used instead)
    swh_storage = unittest.mock.MagicMock(wraps=swh_storage)

    # Add all objects to graph
    swh_graph = unittest.mock.Mock(wraps=GraphClient(nodes=nodes, edges=edges))

    # Cook
    backend = InMemoryVaultBackend()
    cooker = GitBareCooker(
        "revision_gitbare",
        rev2.id,
        backend=backend,
        storage=swh_storage,
        graph=swh_graph,
    )
    cooker.cook()

    # Get bundle
    bundle = backend.fetch("revision_gitbare", rev2.id)

    # Extract bundle and make sure both revisions are in it
    with tempfile.TemporaryDirectory("swh-vault-test-bare") as tempdir:
        with tarfile.open(fileobj=io.BytesIO(bundle)) as tf:
            tf.extractall(tempdir)

        output = subprocess.check_output(
            [
                "git",
                "-C",
                f"{tempdir}/{rev2.swhid()}.git",
                "log",
                "--format=oneline",
                "--decorate=",
            ]
        )

        assert output.decode() == f"{rev2.id.hex()} msg2\n{rev1.id.hex()} msg1\n"

    # Make sure the graph was used instead of swh_storage.revision_log
    swh_graph.visit_nodes.assert_called_once_with(str(rev2.swhid()), edges="rev:rev")
    if last_revision_in_graph:
        swh_storage.revision_log.assert_not_called()
        swh_storage.revision_shortlog.assert_not_called()
    else:
        swh_storage.revision_log.assert_called()



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import os
from typing import Any, Dict

import pytest
import yaml

from swh.core.api.serializers import json_dumps, msgpack_dumps, msgpack_loads
from swh.vault.api.server import (
    VaultServerApp,
    check_config,
    make_app,
    make_app_from_configfile,
)
from swh.vault.tests.test_backend import TEST_HEX_ID


def test_make_app_from_file_missing():
    with pytest.raises(ValueError, match="Missing configuration path."):
        make_app_from_configfile()


def test_make_app_from_file_does_not_exist(tmp_path):
    conf_path = os.path.join(str(tmp_path), "vault-server.yml")
    assert os.path.exists(conf_path) is False

    with pytest.raises(
        ValueError, match=f"Configuration path {conf_path} should exist."
    ):
        make_app_from_configfile(conf_path)


def test_make_app_from_env_variable(swh_vault_config_file):
    """Server initialization happens through env variable when no path is provided

    """
    app = make_app_from_configfile()
    assert app is not None


def test_make_app_from_file(swh_local_vault_config, tmp_path):
    """Server initialization happens trough path if provided

    """
    conf_path = os.path.join(str(tmp_path), "vault-server.yml")
    with open(conf_path, "w") as f:
        f.write(yaml.dump(swh_local_vault_config))

    app = make_app_from_configfile(conf_path)
    assert app is not None


@pytest.fixture
def async_app(swh_local_vault_config: Dict[str, Any],) -> VaultServerApp:
    """Instantiate the vault server application.

    Note: This requires the db setup to run (fixture swh_vault in charge of this)

    """
    return make_app(swh_local_vault_config)


@pytest.fixture
def cli(async_app, aiohttp_client, loop):
    return loop.run_until_complete(aiohttp_client(async_app))


async def test_client_index(cli):
    resp = await cli.get("/")
    assert resp.status == 200


async def test_client_cook_notfound(cli):
    resp = await cli.post(
        "/cook",
        data=json_dumps({"obj_type": "directory", "obj_id": TEST_HEX_ID}),
        headers=[("Content-Type", "application/json")],
    )
    assert resp.status == 400
    content = msgpack_loads(await resp.content.read())
    assert content["type"] == "NotFoundExc"
    assert content["args"] == [f"directory {TEST_HEX_ID} was not found."]


async def test_client_progress_notfound(cli):
    resp = await cli.post(
        "/progress",
        data=json_dumps({"obj_type": "directory", "obj_id": TEST_HEX_ID}),
        headers=[("Content-Type", "application/json")],
    )
    assert resp.status == 400
    content = msgpack_loads(await resp.content.read())
    assert content["type"] == "NotFoundExc"
    assert content["args"] == [f"directory {TEST_HEX_ID} was not found."]


async def test_client_batch_cook_invalid_type(cli):
    resp = await cli.post(
        "/batch_cook",
        data=msgpack_dumps({"batch": [("foobar", [])]}),
        headers={"Content-Type": "application/x-msgpack"},
    )
    assert resp.status == 400
    content = msgpack_loads(await resp.content.read())
    assert content["type"] == "NotFoundExc"
    assert content["args"] == ["foobar is an unknown type."]


async def test_client_batch_progress_notfound(cli):
    resp = await cli.post(
        "/batch_progress",
        data=msgpack_dumps({"batch_id": 1}),
        headers={"Content-Type": "application/x-msgpack"},
    )
    assert resp.status == 400
    content = msgpack_loads(await resp.content.read())
    assert content["type"] == "NotFoundExc"
    assert content["args"] == ["Batch 1 does not exist."]


def test_check_config_missing_vault_configuration() -> None:
    """Irrelevant configuration file path raises"""
    with pytest.raises(ValueError, match="missing 'vault' configuration"):
        check_config({})


def test_check_config_not_local() -> None:
    """Wrong configuration raises"""
    expected_error = (
        "The vault backend can only be started with a 'local' configuration"
    )
    with pytest.raises(EnvironmentError, match=expected_error):
        check_config({"vault": {"cls": "remote"}})


@pytest.mark.parametrize("missing_key", ["storage", "cache", "scheduler"])
def test_check_config_missing_key(missing_key, swh_vault_config) -> None:
    """Any other configuration than 'local' (the default) is rejected"""
    config_ok = {"vault": {"cls": "local", **swh_vault_config}}
    config_ko = copy.deepcopy(config_ok)
    config_ko["vault"].pop(missing_key, None)

    expected_error = f"invalid configuration: missing {missing_key} config entry"
    with pytest.raises(ValueError, match=expected_error):
        check_config(config_ko)


@pytest.mark.parametrize("missing_key", ["storage", "cache", "scheduler"])
def test_check_config_ok(missing_key, swh_vault_config) -> None:
    """Any other configuration than 'local' (the default) is rejected"""
    config_ok = {"vault": {"cls": "local", **swh_vault_config}}
    assert check_config(config_ok) is not None



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import tempfile
from unittest.mock import MagicMock

import click
import click.testing
import pytest

from swh.vault.cli import vault as vault_cli_group
from swh.vault.cookers.base import BaseVaultCooker
from swh.vault.in_memory_backend import InMemoryVaultBackend


def test_cook_unsupported_swhid():
    runner = click.testing.CliRunner()

    result = runner.invoke(vault_cli_group, ["cook", "swh:1:dir:f00b4r", "-"])
    assert isinstance(result.exception, SystemExit)
    assert "expected core SWHID" in result.stdout

    result = runner.invoke(vault_cli_group, ["cook", "swh:1:ori:" + "0" * 40, "-"])
    assert isinstance(result.exception, SystemExit)
    assert "expected core SWHID" in result.stdout

    result = runner.invoke(vault_cli_group, ["cook", "swh:1:cnt:" + "0" * 40, "-"])
    assert isinstance(result.exception, SystemExit)
    assert "No cooker available for CONTENT" in result.stdout


def test_cook_unknown_cooker():
    runner = click.testing.CliRunner()

    result = runner.invoke(
        vault_cli_group,
        ["cook", "swh:1:dir:" + "0" * 40, "-", "--cooker-type", "gitfast"],
    )
    assert isinstance(result.exception, SystemExit)
    assert "do not have a gitfast cooker" in result.stdout

    result = runner.invoke(vault_cli_group, ["cook", "swh:1:rev:" + "0" * 40, "-"])
    assert isinstance(result.exception, SystemExit)
    assert "explicit --cooker-type" in result.stdout


@pytest.mark.parametrize(
    "obj_type,cooker_name_suffix,swhid_type",
    [("directory", "", "dir"), ("revision", "gitfast", "rev"),],
)
def test_cook_directory(obj_type, cooker_name_suffix, swhid_type, mocker):
    storage = object()
    mocker.patch("swh.storage.get_storage", return_value=storage)

    backend = MagicMock(spec=InMemoryVaultBackend)
    backend.fetch.return_value = b"bundle content"
    mocker.patch(
        "swh.vault.in_memory_backend.InMemoryVaultBackend", return_value=backend
    )

    cooker = MagicMock(spec=BaseVaultCooker)
    cooker_cls = MagicMock(return_value=cooker)
    mocker.patch("swh.vault.cookers.get_cooker_cls", return_value=cooker_cls)

    runner = click.testing.CliRunner()

    with tempfile.NamedTemporaryFile("a", suffix=".yml") as config_fd:
        config_fd.write('{"storage": {}}')
        config_fd.seek(0)
        if cooker_name_suffix:
            result = runner.invoke(
                vault_cli_group,
                [
                    "cook",
                    f"swh:1:{swhid_type}:{'0'*40}",
                    "-",
                    "-C",
                    config_fd.name,
                    "--cooker-type",
                    cooker_name_suffix,
                ],
            )
        else:
            result = runner.invoke(
                vault_cli_group,
                ["cook", f"swh:1:{swhid_type}:{'0'*40}", "-", "-C", config_fd.name],
            )

    if result.exception is not None:
        raise result.exception

    cooker_cls.assert_called_once_with(
        obj_type=f"{obj_type}_{cooker_name_suffix}" if cooker_name_suffix else obj_type,
        obj_id=b"\x00" * 20,
        backend=backend,
        storage=storage,
        graph=None,
        objstorage=None,
    )
    cooker.cook.assert_called_once_with()

    assert result.stdout_bytes == b"bundle content"



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import contextlib
import datetime
import glob
import gzip
import io
import os
import pathlib
import shutil
import subprocess
import tarfile
import tempfile
import unittest
import unittest.mock

import dulwich.fastexport
import dulwich.index
import dulwich.objects
import dulwich.porcelain
import dulwich.repo
import pytest

from swh.loader.git.from_disk import GitLoaderFromDisk
from swh.model import from_disk, hashutil
from swh.model.model import (
    Directory,
    DirectoryEntry,
    Person,
    Revision,
    RevisionType,
    TimestampWithTimezone,
)
from swh.vault.cookers import DirectoryCooker, GitBareCooker, RevisionGitfastCooker
from swh.vault.tests.vault_testing import hash_content
from swh.vault.to_disk import HIDDEN_MESSAGE, SKIPPED_MESSAGE


class TestRepo:
    """A tiny context manager for a test git repository, with some utility
    functions to perform basic git stuff.
    """

    def __init__(self, repo_dir=None):
        self.repo_dir = repo_dir

    def __enter__(self):
        if self.repo_dir:
            self.tmp_dir = None
            self.repo = dulwich.repo.Repo(self.repo_dir)
        else:
            self.tmp_dir = tempfile.TemporaryDirectory(prefix="tmp-vault-repo-")
            self.repo_dir = self.tmp_dir.__enter__()
            self.repo = dulwich.repo.Repo.init(self.repo_dir)
        self.author_name = b"Test Author"
        self.author_email = b"test@softwareheritage.org"
        self.author = b"%s <%s>" % (self.author_name, self.author_email)
        self.base_date = 258244200
        self.counter = 0
        return pathlib.Path(self.repo_dir)

    def __exit__(self, exc, value, tb):
        if self.tmp_dir is not None:
            self.tmp_dir.__exit__(exc, value, tb)
            self.repo_dir = None

    def checkout(self, rev_sha):
        rev = self.repo[rev_sha]
        dulwich.index.build_index_from_tree(
            self.repo_dir, self.repo.index_path(), self.repo.object_store, rev.tree
        )

    def git_shell(self, *cmd, stdout=subprocess.DEVNULL, **kwargs):
        name = self.author_name
        email = self.author_email
        date = "%d +0000" % (self.base_date + self.counter)
        env = {
            # Set git commit format
            "GIT_AUTHOR_NAME": name,
            "GIT_AUTHOR_EMAIL": email,
            "GIT_AUTHOR_DATE": date,
            "GIT_COMMITTER_NAME": name,
            "GIT_COMMITTER_EMAIL": email,
            "GIT_COMMITTER_DATE": date,
            # Ignore all the system-wide and user configurations
            "GIT_CONFIG_NOSYSTEM": "1",
            "HOME": str(self.tmp_dir),
            "XDG_CONFIG_HOME": str(self.tmp_dir),
        }
        kwargs.setdefault("env", {}).update(env)

        subprocess.check_call(
            ("git", "-C", self.repo_dir) + cmd, stdout=stdout, **kwargs
        )

    def commit(self, message="Commit test\n", ref=b"HEAD"):
        """Commit the current working tree in a new commit with message on
        the branch 'ref'.

        At the end of the commit, the reference should stay the same
        and the index should be clean.

        """
        paths = [
            os.path.relpath(path, self.repo_dir)
            for path in glob.glob(self.repo_dir + "/**/*", recursive=True)
        ]
        self.repo.stage(paths)
        message = message.encode() + b"\n"
        ret = self.repo.do_commit(
            message=message,
            committer=self.author,
            commit_timestamp=self.base_date + self.counter,
            commit_timezone=0,
            ref=ref,
        )
        self.counter += 1

        # committing on another branch leaves
        # dangling files in index
        if ref != b"HEAD":
            # XXX this should work (but does not)
            # dulwich.porcelain.reset(self.repo, 'hard')
            self.git_shell("reset", "--hard", "HEAD")
        return ret

    def merge(self, parent_sha_list, message="Merge branches."):
        self.git_shell(
            "merge",
            "--allow-unrelated-histories",
            "-m",
            message,
            *[p.decode() for p in parent_sha_list],
        )
        self.counter += 1
        return self.repo.refs[b"HEAD"]

    def print_debug_graph(self, reflog=False):
        args = ["log", "--all", "--graph", "--decorate"]
        if reflog:
            args.append("--reflog")
        self.git_shell(*args, stdout=None)


@pytest.fixture
def git_loader(swh_storage,):
    """Instantiate a Git Loader using the storage instance as storage.

    """

    def _create_loader(directory):
        return GitLoaderFromDisk(
            swh_storage,
            "fake_origin",
            directory=directory,
            visit_date=datetime.datetime.now(datetime.timezone.utc),
        )

    return _create_loader


@contextlib.contextmanager
def cook_extract_directory_dircooker(storage, obj_id, fsck=True):
    """Context manager that cooks a directory and extract it."""
    backend = unittest.mock.MagicMock()
    backend.storage = storage
    cooker = DirectoryCooker("directory", obj_id, backend=backend, storage=storage)
    cooker.fileobj = io.BytesIO()
    assert cooker.check_exists()
    cooker.prepare_bundle()
    cooker.fileobj.seek(0)
    with tempfile.TemporaryDirectory(prefix="tmp-vault-extract-") as td:
        with tarfile.open(fileobj=cooker.fileobj, mode="r") as tar:
            tar.extractall(td)
        yield pathlib.Path(td) / hashutil.hash_to_hex(obj_id)
    cooker.storage = None


@contextlib.contextmanager
def cook_extract_directory_gitfast(storage, obj_id, fsck=True):
    """Context manager that cooks a revision containing a directory and extract it,
    using RevisionGitfastCooker"""
    test_repo = TestRepo()
    with test_repo as p:
        date = TimestampWithTimezone.from_datetime(
            datetime.datetime.now(datetime.timezone.utc)
        )
        revision = Revision(
            directory=obj_id,
            message=b"dummy message",
            author=Person.from_fullname(b"someone"),
            committer=Person.from_fullname(b"someone"),
            date=date,
            committer_date=date,
            type=RevisionType.GIT,
            synthetic=False,
        )
        storage.revision_add([revision])

    with cook_stream_revision_gitfast(storage, revision.id) as stream, test_repo as p:
        processor = dulwich.fastexport.GitImportProcessor(test_repo.repo)
        processor.import_stream(stream)
        test_repo.checkout(b"HEAD")
        shutil.rmtree(p / ".git")
        yield p


@contextlib.contextmanager
def cook_extract_directory_git_bare(
    storage, obj_id, fsck=True, direct_objstorage=False
):
    """Context manager that cooks a revision and extract it,
    using GitBareCooker"""
    backend = unittest.mock.MagicMock()
    backend.storage = storage

    # Cook the object
    cooker = GitBareCooker(
        "directory",
        obj_id,
        backend=backend,
        storage=storage,
        objstorage=storage.objstorage if direct_objstorage else None,
    )
    cooker.use_fsck = fsck  # Some tests try edge-cases that git-fsck rejects
    cooker.fileobj = io.BytesIO()
    assert cooker.check_exists()
    cooker.prepare_bundle()
    cooker.fileobj.seek(0)

    # Extract it
    with tempfile.TemporaryDirectory(prefix="tmp-vault-extract-") as td:
        with tarfile.open(fileobj=cooker.fileobj, mode="r") as tar:
            tar.extractall(td)

        # Clone it with Dulwich
        with tempfile.TemporaryDirectory(prefix="tmp-vault-clone-") as clone_dir:
            clone_dir = pathlib.Path(clone_dir)
            subprocess.check_call(
                [
                    "git",
                    "clone",
                    os.path.join(td, f"swh:1:dir:{obj_id.hex()}.git"),
                    clone_dir,
                ]
            )
            shutil.rmtree(clone_dir / ".git")
            yield clone_dir


@pytest.fixture(
    scope="module",
    params=[
        cook_extract_directory_dircooker,
        cook_extract_directory_gitfast,
        cook_extract_directory_git_bare,
    ],
)
def cook_extract_directory(request):
    """A fixture that is instantiated as either cook_extract_directory_dircooker or
    cook_extract_directory_git_bare."""
    return request.param


@contextlib.contextmanager
def cook_stream_revision_gitfast(storage, obj_id):
    """Context manager that cooks a revision and stream its fastexport."""
    backend = unittest.mock.MagicMock()
    backend.storage = storage
    cooker = RevisionGitfastCooker(
        "revision_gitfast", obj_id, backend=backend, storage=storage
    )
    cooker.fileobj = io.BytesIO()
    assert cooker.check_exists()
    cooker.prepare_bundle()
    cooker.fileobj.seek(0)
    fastexport_stream = gzip.GzipFile(fileobj=cooker.fileobj)
    yield fastexport_stream
    cooker.storage = None


@contextlib.contextmanager
def cook_extract_revision_gitfast(storage, obj_id, fsck=True):
    """Context manager that cooks a revision and extract it,
    using RevisionGitfastCooker"""
    test_repo = TestRepo()
    with cook_stream_revision_gitfast(storage, obj_id) as stream, test_repo as p:
        processor = dulwich.fastexport.GitImportProcessor(test_repo.repo)
        processor.import_stream(stream)
        yield test_repo, p


@contextlib.contextmanager
def cook_extract_revision_git_bare(storage, obj_id, fsck=True):
    """Context manager that cooks a revision and extract it,
    using GitBareCooker"""
    backend = unittest.mock.MagicMock()
    backend.storage = storage

    # Cook the object
    cooker = GitBareCooker("revision", obj_id, backend=backend, storage=storage)
    cooker.use_fsck = fsck  # Some tests try edge-cases that git-fsck rejects
    cooker.fileobj = io.BytesIO()
    assert cooker.check_exists()
    cooker.prepare_bundle()
    cooker.fileobj.seek(0)

    # Extract it
    with tempfile.TemporaryDirectory(prefix="tmp-vault-extract-") as td:
        with tarfile.open(fileobj=cooker.fileobj, mode="r") as tar:
            tar.extractall(td)

        # Clone it with Dulwich
        with tempfile.TemporaryDirectory(prefix="tmp-vault-clone-") as clone_dir:
            clone_dir = pathlib.Path(clone_dir)
            subprocess.check_call(
                [
                    "git",
                    "clone",
                    os.path.join(td, f"swh:1:rev:{obj_id.hex()}.git"),
                    clone_dir,
                ]
            )
            test_repo = TestRepo(clone_dir)
            with test_repo:
                yield test_repo, clone_dir


@pytest.fixture(
    scope="module",
    params=[cook_extract_revision_gitfast, cook_extract_revision_git_bare],
)
def cook_extract_revision(request):
    """A fixture that is instantiated as either cook_extract_revision_gitfast or
    cook_extract_revision_git_bare."""
    return request.param


TEST_CONTENT = (
    "   test content\n" "and unicode \N{BLACK HEART SUIT}\n" " and trailing spaces   "
)
TEST_EXECUTABLE = b"\x42\x40\x00\x00\x05"


class TestDirectoryCooker:
    def test_directory_simple(self, git_loader, cook_extract_directory):
        repo = TestRepo()
        with repo as rp:
            (rp / "file").write_text(TEST_CONTENT)
            (rp / "executable").write_bytes(TEST_EXECUTABLE)
            (rp / "executable").chmod(0o755)
            (rp / "link").symlink_to("file")
            (rp / "dir1/dir2").mkdir(parents=True)
            (rp / "dir1/dir2/file").write_text(TEST_CONTENT)
            c = repo.commit()
            loader = git_loader(str(rp))
            loader.load()

            obj_id_hex = repo.repo[c].tree.decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)

        with cook_extract_directory(loader.storage, obj_id) as p:
            assert (p / "file").stat().st_mode == 0o100644
            assert (p / "file").read_text() == TEST_CONTENT
            assert (p / "executable").stat().st_mode == 0o100755
            assert (p / "executable").read_bytes() == TEST_EXECUTABLE
            assert (p / "link").is_symlink()
            assert os.readlink(str(p / "link")) == "file"
            assert (p / "dir1/dir2/file").stat().st_mode == 0o100644
            assert (p / "dir1/dir2/file").read_text() == TEST_CONTENT

            directory = from_disk.Directory.from_disk(path=bytes(p))
            assert obj_id_hex == hashutil.hash_to_hex(directory.hash)

    def test_directory_filtered_objects(self, git_loader, cook_extract_directory):
        repo = TestRepo()
        with repo as rp:
            file_1, id_1 = hash_content(b"test1")
            file_2, id_2 = hash_content(b"test2")
            file_3, id_3 = hash_content(b"test3")

            (rp / "file").write_bytes(file_1)
            (rp / "hidden_file").write_bytes(file_2)
            (rp / "absent_file").write_bytes(file_3)

            c = repo.commit()
            loader = git_loader(str(rp))
            loader.load()

            obj_id_hex = repo.repo[c].tree.decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)

        # FIXME: storage.content_update() should be changed to allow things
        # like that
        with loader.storage.get_db().transaction() as cur:
            cur.execute(
                """update content set status = 'visible'
                           where sha1 = %s""",
                (id_1,),
            )
            cur.execute(
                """update content set status = 'hidden'
                           where sha1 = %s""",
                (id_2,),
            )

            cur.execute(
                """
                insert into skipped_content
                    (sha1, sha1_git, sha256, blake2s256, length, reason)
                select sha1, sha1_git, sha256, blake2s256, length, 'no reason'
                from content
                where sha1 = %s
                """,
                (id_3,),
            )

            cur.execute("delete from content where sha1 = %s", (id_3,))

        with cook_extract_directory(loader.storage, obj_id) as p:
            assert (p / "file").read_bytes() == b"test1"
            assert (p / "hidden_file").read_bytes() == HIDDEN_MESSAGE
            assert (p / "absent_file").read_bytes() == SKIPPED_MESSAGE

    def test_directory_bogus_perms(self, git_loader, cook_extract_directory):
        # Some early git repositories have 664/775 permissions... let's check
        # if all the weird modes are properly normalized in the directory
        # cooker.
        repo = TestRepo()
        with repo as rp:
            (rp / "file").write_text(TEST_CONTENT)
            (rp / "file").chmod(0o664)
            (rp / "executable").write_bytes(TEST_EXECUTABLE)
            (rp / "executable").chmod(0o775)
            (rp / "wat").write_text(TEST_CONTENT)
            (rp / "wat").chmod(0o604)

            # Disable mode cleanup
            with unittest.mock.patch("dulwich.index.cleanup_mode", lambda mode: mode):
                c = repo.commit()

            # Make sure Dulwich didn't normalize the permissions itself.
            # (if it did, then the test can't check the cooker normalized them)
            tree_id = repo.repo[c].tree
            assert {entry.mode for entry in repo.repo[tree_id].items()} == {
                0o100775,
                0o100664,
                0o100604,
            }

            # Disable mode checks
            with unittest.mock.patch("dulwich.objects.Tree.check", lambda self: None):
                loader = git_loader(str(rp))
                loader.load()

            # Make sure swh-loader didn't normalize them either
            dir_entries = loader.storage.directory_ls(hashutil.bytehex_to_hash(tree_id))
            assert {entry["perms"] for entry in dir_entries} == {
                0o100664,
                0o100775,
                0o100604,
            }

            obj_id_hex = repo.repo[c].tree.decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)

        with cook_extract_directory(loader.storage, obj_id) as p:
            assert (p / "file").stat().st_mode == 0o100644
            assert (p / "executable").stat().st_mode == 0o100755
            assert (p / "wat").stat().st_mode == 0o100644

    @pytest.mark.parametrize("direct_objstorage", [True, False])
    def test_directory_objstorage(
        self, swh_storage, git_loader, mocker, direct_objstorage
    ):
        """Like test_directory_simple, but using swh_objstorage directly, without
        going through swh_storage.content_get_data()"""
        repo = TestRepo()
        with repo as rp:
            (rp / "file").write_text(TEST_CONTENT)
            (rp / "executable").write_bytes(TEST_EXECUTABLE)
            (rp / "executable").chmod(0o755)
            (rp / "link").symlink_to("file")
            (rp / "dir1/dir2").mkdir(parents=True)
            (rp / "dir1/dir2/file").write_text(TEST_CONTENT)
            c = repo.commit()
            loader = git_loader(str(rp))
            loader.load()

            obj_id_hex = repo.repo[c].tree.decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)

        # Set-up spies
        storage_content_get_data = mocker.patch.object(
            swh_storage, "content_get_data", wraps=swh_storage.content_get_data
        )
        objstorage_content_batch = mocker.patch.object(
            swh_storage.objstorage, "get_batch", wraps=swh_storage.objstorage.get_batch
        )

        with cook_extract_directory_git_bare(
            loader.storage, obj_id, direct_objstorage=direct_objstorage
        ) as p:
            assert (p / "file").stat().st_mode == 0o100644
            assert (p / "file").read_text() == TEST_CONTENT
            assert (p / "executable").stat().st_mode == 0o100755
            assert (p / "executable").read_bytes() == TEST_EXECUTABLE
            assert (p / "link").is_symlink()
            assert os.readlink(str(p / "link")) == "file"
            assert (p / "dir1/dir2/file").stat().st_mode == 0o100644
            assert (p / "dir1/dir2/file").read_text() == TEST_CONTENT

            directory = from_disk.Directory.from_disk(path=bytes(p))
            assert obj_id_hex == hashutil.hash_to_hex(directory.hash)

        if direct_objstorage:
            storage_content_get_data.assert_not_called()
            objstorage_content_batch.assert_called()
        else:
            storage_content_get_data.assert_called()
            objstorage_content_batch.assert_not_called()

    def test_directory_revision_data(self, swh_storage):
        target_rev = "0e8a3ad980ec179856012b7eecf4327e99cd44cd"

        dir = Directory(
            entries=(
                DirectoryEntry(
                    name=b"submodule",
                    type="rev",
                    target=hashutil.hash_to_bytes(target_rev),
                    perms=0o100644,
                ),
            ),
        )
        swh_storage.directory_add([dir])

        with cook_extract_directory_dircooker(swh_storage, dir.id, fsck=False) as p:
            assert (p / "submodule").is_symlink()
            assert os.readlink(str(p / "submodule")) == target_rev


class TestRevisionCooker:
    def test_revision_simple(self, git_loader, cook_extract_revision):
        #
        #     1--2--3--4--5--6--7
        #
        repo = TestRepo()
        with repo as rp:
            (rp / "file1").write_text(TEST_CONTENT)
            repo.commit("add file1")
            (rp / "file2").write_text(TEST_CONTENT)
            repo.commit("add file2")
            (rp / "dir1/dir2").mkdir(parents=True)
            (rp / "dir1/dir2/file").write_text(TEST_CONTENT)
            repo.commit("add dir1/dir2/file")
            (rp / "bin1").write_bytes(TEST_EXECUTABLE)
            (rp / "bin1").chmod(0o755)
            repo.commit("add bin1")
            (rp / "link1").symlink_to("file1")
            repo.commit("link link1 to file1")
            (rp / "file2").unlink()
            repo.commit("remove file2")
            (rp / "bin1").rename(rp / "bin")
            repo.commit("rename bin1 to bin")
            loader = git_loader(str(rp))
            loader.load()
            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)

        with cook_extract_revision(loader.storage, obj_id) as (ert, p):
            ert.checkout(b"HEAD")
            assert (p / "file1").stat().st_mode == 0o100644
            assert (p / "file1").read_text() == TEST_CONTENT
            assert (p / "link1").is_symlink()
            assert os.readlink(str(p / "link1")) == "file1"
            assert (p / "bin").stat().st_mode == 0o100755
            assert (p / "bin").read_bytes() == TEST_EXECUTABLE
            assert (p / "dir1/dir2/file").read_text() == TEST_CONTENT
            assert (p / "dir1/dir2/file").stat().st_mode == 0o100644
            assert ert.repo.refs[b"HEAD"].decode() == obj_id_hex

    def test_revision_two_roots(self, git_loader, cook_extract_revision):
        #
        #    1----3---4
        #        /
        #   2----
        #
        repo = TestRepo()
        with repo as rp:
            (rp / "file1").write_text(TEST_CONTENT)
            c1 = repo.commit("Add file1")
            del repo.repo.refs[b"refs/heads/master"]  # git update-ref -d HEAD
            (rp / "file2").write_text(TEST_CONTENT)
            repo.commit("Add file2")
            repo.merge([c1])
            (rp / "file3").write_text(TEST_CONTENT)
            repo.commit("add file3")
            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            loader = git_loader(str(rp))
            loader.load()

        with cook_extract_revision(loader.storage, obj_id) as (ert, p):
            assert ert.repo.refs[b"HEAD"].decode() == obj_id_hex

    def test_revision_two_double_fork_merge(self, git_loader, cook_extract_revision):
        #
        #     2---4---6
        #    /   /   /
        #   1---3---5
        #
        repo = TestRepo()
        with repo as rp:
            (rp / "file1").write_text(TEST_CONTENT)
            c1 = repo.commit("Add file1")
            repo.repo.refs[b"refs/heads/c1"] = c1

            (rp / "file2").write_text(TEST_CONTENT)
            repo.commit("Add file2")

            (rp / "file3").write_text(TEST_CONTENT)
            c3 = repo.commit("Add file3", ref=b"refs/heads/c1")
            repo.repo.refs[b"refs/heads/c3"] = c3

            repo.merge([c3])

            (rp / "file5").write_text(TEST_CONTENT)
            c5 = repo.commit("Add file3", ref=b"refs/heads/c3")

            repo.merge([c5])

            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            loader = git_loader(str(rp))
            loader.load()

        with cook_extract_revision(loader.storage, obj_id) as (ert, p):
            assert ert.repo.refs[b"HEAD"].decode() == obj_id_hex

    def test_revision_triple_merge(self, git_loader, cook_extract_revision):
        #
        #       .---.---5
        #      /   /   /
        #     2   3   4
        #    /   /   /
        #   1---.---.
        #
        repo = TestRepo()
        with repo as rp:
            (rp / "file1").write_text(TEST_CONTENT)
            c1 = repo.commit("Commit 1")
            repo.repo.refs[b"refs/heads/b1"] = c1
            repo.repo.refs[b"refs/heads/b2"] = c1

            repo.commit("Commit 2")
            c3 = repo.commit("Commit 3", ref=b"refs/heads/b1")
            c4 = repo.commit("Commit 4", ref=b"refs/heads/b2")
            repo.merge([c3, c4])

            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            loader = git_loader(str(rp))
            loader.load()

        with cook_extract_revision(loader.storage, obj_id) as (ert, p):
            assert ert.repo.refs[b"HEAD"].decode() == obj_id_hex

    def test_revision_filtered_objects(self, git_loader, cook_extract_revision):
        repo = TestRepo()
        with repo as rp:
            file_1, id_1 = hash_content(b"test1")
            file_2, id_2 = hash_content(b"test2")
            file_3, id_3 = hash_content(b"test3")

            (rp / "file").write_bytes(file_1)
            (rp / "hidden_file").write_bytes(file_2)
            (rp / "absent_file").write_bytes(file_3)

            repo.commit()
            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            loader = git_loader(str(rp))
            loader.load()

        # FIXME: storage.content_update() should be changed to allow things
        # like that
        with loader.storage.get_db().transaction() as cur:
            cur.execute(
                """update content set status = 'visible'
                           where sha1 = %s""",
                (id_1,),
            )
            cur.execute(
                """update content set status = 'hidden'
                           where sha1 = %s""",
                (id_2,),
            )

            cur.execute(
                """
                insert into skipped_content
                    (sha1, sha1_git, sha256, blake2s256, length, reason)
                select sha1, sha1_git, sha256, blake2s256, length, 'no reason'
                from content
                where sha1 = %s
                """,
                (id_3,),
            )

            cur.execute("delete from content where sha1 = %s", (id_3,))

        with cook_extract_revision(loader.storage, obj_id) as (ert, p):
            ert.checkout(b"HEAD")
            assert (p / "file").read_bytes() == b"test1"
            assert (p / "hidden_file").read_bytes() == HIDDEN_MESSAGE
            assert (p / "absent_file").read_bytes() == SKIPPED_MESSAGE

    def test_revision_null_fields(self, git_loader, cook_extract_revision):
        # Our schema doesn't enforce a lot of non-null revision fields. We need
        # to check these cases don't break the cooker.
        repo = TestRepo()
        with repo as rp:
            (rp / "file").write_text(TEST_CONTENT)
            c = repo.commit("initial commit")
            loader = git_loader(str(rp))
            loader.load()
            repo.repo.refs[b"HEAD"].decode()
            dir_id_hex = repo.repo[c].tree.decode()
            dir_id = hashutil.hash_to_bytes(dir_id_hex)

        test_revision = Revision(
            message=b"",
            author=Person(name=None, email=None, fullname=b""),
            date=None,
            committer=Person(name=None, email=None, fullname=b""),
            committer_date=None,
            parents=(),
            type=RevisionType.GIT,
            directory=dir_id,
            metadata={},
            synthetic=True,
        )

        storage = loader.storage
        storage.revision_add([test_revision])

        with cook_extract_revision(storage, test_revision.id, fsck=False) as (ert, p):
            ert.checkout(b"HEAD")
            assert (p / "file").stat().st_mode == 0o100644

    def test_revision_revision_data(self, swh_storage):
        target_rev = "0e8a3ad980ec179856012b7eecf4327e99cd44cd"

        dir = Directory(
            entries=(
                DirectoryEntry(
                    name=b"submodule",
                    type="rev",
                    target=hashutil.hash_to_bytes(target_rev),
                    perms=0o100644,
                ),
            ),
        )
        swh_storage.directory_add([dir])

        rev = Revision(
            message=b"",
            author=Person(name=None, email=None, fullname=b""),
            date=None,
            committer=Person(name=None, email=None, fullname=b""),
            committer_date=None,
            parents=(),
            type=RevisionType.GIT,
            directory=dir.id,
            metadata={},
            synthetic=True,
        )
        swh_storage.revision_add([rev])

        with cook_stream_revision_gitfast(swh_storage, rev.id) as stream:
            pattern = "M 160000 {} submodule".format(target_rev).encode()
            assert pattern in stream.read()



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from typing import Any, Dict

import pkg_resources.extern.packaging.version
import pytest
import yaml

from swh.core.db.pytest_plugin import postgresql_fact
from swh.storage.tests import SQL_DIR as STORAGE_SQL_DIR
import swh.vault
from swh.vault import get_vault

os.environ["LC_ALL"] = "C.UTF-8"

pytest_v = pkg_resources.get_distribution("pytest").parsed_version
if pytest_v < pkg_resources.extern.packaging.version.parse("3.9"):

    @pytest.fixture
    def tmp_path():
        import pathlib
        import tempfile

        with tempfile.TemporaryDirectory() as tmpdir:
            yield pathlib.Path(tmpdir)


VAULT_SQL_DIR = os.path.join(os.path.dirname(swh.vault.__file__), "sql")


postgres_vault = postgresql_fact(
    "postgresql_proc", dbname="vault", dump_files=f"{VAULT_SQL_DIR}/*.sql"
)
postgres_storage = postgresql_fact(
    "postgresql_proc", dbname="storage", dump_files=f"{STORAGE_SQL_DIR}/*.sql"
)


@pytest.fixture
def swh_vault_config(postgres_vault, postgres_storage, tmp_path) -> Dict[str, Any]:
    tmp_path = str(tmp_path)
    return {
        "db": postgres_vault.dsn,
        "storage": {
            "cls": "local",
            "db": postgres_storage.dsn,
            "objstorage": {
                "cls": "pathslicing",
                "args": {"root": tmp_path, "slicing": "0:1/1:5",},
            },
        },
        "cache": {
            "cls": "pathslicing",
            "args": {"root": tmp_path, "slicing": "0:1/1:5", "allow_delete": True},
        },
        "scheduler": {"cls": "remote", "url": "http://swh-scheduler:5008",},
    }


@pytest.fixture
def swh_local_vault_config(swh_vault_config: Dict[str, Any]) -> Dict[str, Any]:
    return {
        "vault": {"cls": "local", **swh_vault_config},
        "client_max_size": 1024 ** 3,
    }


@pytest.fixture
def swh_vault_config_file(swh_local_vault_config, monkeypatch, tmp_path):
    conf_path = os.path.join(str(tmp_path), "vault-server.yml")
    with open(conf_path, "w") as f:
        f.write(yaml.dump(swh_local_vault_config))
    monkeypatch.setenv("SWH_CONFIG_FILENAME", conf_path)
    return conf_path


@pytest.fixture
def swh_vault(swh_vault_config):
    return get_vault("local", **swh_vault_config)


@pytest.fixture
def swh_storage(swh_vault):
    return swh_vault.storage



# Copyright (C) 2017-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

import os
from typing import Any, Dict

from swh.core.config import load_named_config
from swh.core.config import read as read_config
from swh.storage import get_storage
from swh.vault import get_vault
from swh.vault.cookers.base import DEFAULT_CONFIG, DEFAULT_CONFIG_PATH
from swh.vault.cookers.directory import DirectoryCooker
from swh.vault.cookers.git_bare import GitBareCooker
from swh.vault.cookers.revision_flat import RevisionFlatCooker
from swh.vault.cookers.revision_gitfast import RevisionGitfastCooker

COOKER_TYPES = {
    "directory": DirectoryCooker,
    "revision_flat": RevisionFlatCooker,
    "revision_gitfast": RevisionGitfastCooker,
    "revision_git_bare": GitBareCooker,
    "directory_git_bare": GitBareCooker,
}


def get_cooker_cls(obj_type):
    return COOKER_TYPES[obj_type]


def check_config(cfg: Dict[str, Any]) -> Dict[str, Any]:
    """Ensure the configuration is ok to run a vault worker, and propagate defaults

    Raises:
        EnvironmentError if the configuration is not for remote instance
        ValueError if one of the following keys is missing: vault, storage

    Returns:
        New configuration dict to instantiate a vault worker instance

    """
    cfg = cfg.copy()

    if "vault" not in cfg:
        raise ValueError("missing 'vault' configuration")

    vcfg = cfg["vault"]
    if vcfg["cls"] != "remote":
        raise EnvironmentError(
            "This vault backend can only be a 'remote' configuration"
        )

    # TODO: Soft-deprecation of args key. Remove when ready.
    vcfg.update(vcfg.get("args", {}))

    # Default to top-level value if any
    if "storage" not in vcfg:
        vcfg["storage"] = cfg.get("storage")

    if not vcfg.get("storage"):
        raise ValueError("invalid configuration: missing 'storage' config entry.")

    return cfg


def get_cooker(obj_type: str, obj_id: str):
    """Instantiate a cooker class of type obj_type.

    Returns:
        Cooker class in charge of cooking the obj_type with id obj_id.

    Raises:
        ValueError in case of a missing top-level vault key configuration or a storage
          key.
        EnvironmentError in case the vault configuration reference a non remote class.

    """
    if "SWH_CONFIG_FILENAME" in os.environ:
        cfg = read_config(os.environ["SWH_CONFIG_FILENAME"], DEFAULT_CONFIG)
    else:
        cfg = load_named_config(DEFAULT_CONFIG_PATH, DEFAULT_CONFIG)
    cooker_cls = get_cooker_cls(obj_type)

    cfg = check_config(cfg)
    vcfg = cfg["vault"]

    storage = get_storage(**vcfg.pop("storage"))
    backend = get_vault(**vcfg)

    try:
        from swh.graph.client import RemoteGraphClient  # optional dependency

        graph = RemoteGraphClient(**vcfg["graph"]) if vcfg.get("graph") else None
    except ModuleNotFoundError:
        if vcfg.get("graph"):
            raise EnvironmentError(
                "Graph configuration required but module is not installed."
            )
        else:
            graph = None

    return cooker_cls(
        obj_type,
        obj_id,
        backend=backend,
        storage=storage,
        graph=graph,
        max_bundle_size=cfg["max_bundle_size"],
    )



# Copyright (C) 2016-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

import asyncio
import os
from typing import Any, Dict, Optional

import aiohttp.web

from swh.core.api.asynchronous import RPCServerApp
from swh.core.config import config_basepath, merge_configs, read_raw_config
from swh.vault import get_vault as get_swhvault
from swh.vault.backend import NotFoundExc
from swh.vault.interface import VaultInterface

# do not define default services here
DEFAULT_CONFIG = {
    "client_max_size": 1024 ** 3,
}


vault = None
app = None


def get_vault(config: Optional[Dict[str, Any]] = None) -> VaultInterface:
    global vault
    if not vault:
        assert config is not None
        vault = get_swhvault(**config)
    return vault


class VaultServerApp(RPCServerApp):
    client_exception_classes = (NotFoundExc,)


@asyncio.coroutine
def index(request):
    return aiohttp.web.Response(body="SWH Vault API server")


def check_config(cfg: Dict[str, Any]) -> Dict[str, Any]:
    """Ensure the configuration is ok to run a local vault server, and propagate defaults.

    Raises:
        EnvironmentError if the configuration is not for local instance
        ValueError if one of the following keys is missing: vault, cache, storage,
        scheduler

    Returns:
        New configuration dict to instantiate a local vault server instance.

    """
    cfg = cfg.copy()

    if "vault" not in cfg:
        raise ValueError("missing 'vault' configuration")

    vcfg = cfg["vault"]
    if vcfg["cls"] != "local":
        raise EnvironmentError(
            "The vault backend can only be started with a 'local' configuration",
        )

    # TODO: Soft-deprecation of args key. Remove when ready.
    vcfg.update(vcfg.get("args", {}))

    # Default to top-level value if any
    if "cache" not in vcfg:
        vcfg["cache"] = cfg.get("cache")
    if "storage" not in vcfg:
        vcfg["storage"] = cfg.get("storage")
    if "scheduler" not in vcfg:
        vcfg["scheduler"] = cfg.get("scheduler")
    if "client_max_size" not in vcfg:
        vcfg["client_max_size"] = cfg.get("client_max_size")

    for key in ("cache", "storage", "scheduler"):
        if not vcfg.get(key):
            raise ValueError(f"invalid configuration: missing {key} config entry.")

    return vcfg


def make_app(config: Dict[str, Any]) -> VaultServerApp:
    """Ensure the configuration is ok, then instantiate the server application

    """
    config = check_config(config)
    app = VaultServerApp(
        __name__,
        backend_class=VaultInterface,
        backend_factory=lambda: get_vault(config),
        client_max_size=config["client_max_size"],
    )
    app.router.add_route("GET", "/", index)
    return app


def make_app_from_configfile(
    config_path: Optional[str] = None, **kwargs
) -> VaultServerApp:
    """Load and check configuration if ok, then instantiate (once) a vault server
       application.

    """
    global app
    if not app:
        config_path = os.environ.get("SWH_CONFIG_FILENAME", config_path)
        if not config_path:
            raise ValueError("Missing configuration path.")
        if not os.path.isfile(config_path):
            raise ValueError(f"Configuration path {config_path} should exist.")

        app_config = read_raw_config(config_basepath(config_path))
        app_config = merge_configs(DEFAULT_CONFIG, app_config)
        app = make_app(app_config)

    return app


if __name__ == "__main__":
    print("Deprecated. Use swh-vault ")



# Copyright (C) 2016-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
import io
import logging
from typing import Optional

from psycopg2.extensions import QueryCanceledError

from swh.model import hashutil

MAX_BUNDLE_SIZE = 2 ** 29  # 512 MiB
DEFAULT_CONFIG_PATH = "vault/cooker"
DEFAULT_CONFIG = {
    "max_bundle_size": ("int", MAX_BUNDLE_SIZE),
}


class PolicyError(Exception):
    """Raised when the bundle violates the cooking policy."""

    pass


class BundleTooLargeError(PolicyError):
    """Raised when the bundle is too large to be cooked."""

    pass


class BytesIOBundleSizeLimit(io.BytesIO):
    def __init__(self, *args, size_limit=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.size_limit = size_limit

    def write(self, chunk):
        if (
            self.size_limit is not None
            and self.getbuffer().nbytes + len(chunk) > self.size_limit
        ):
            raise BundleTooLargeError(
                "The requested bundle exceeds the maximum allowed "
                "size of {} bytes.".format(self.size_limit)
            )
        return super().write(chunk)


class BaseVaultCooker(metaclass=abc.ABCMeta):
    """Abstract base class for the vault's bundle creators

    This class describes a common API for the cookers.

    To define a new cooker, inherit from this class and override:
    - CACHE_TYPE_KEY: key to use for the bundle to reference in cache
    - def cook(): cook the object into a bundle
    """

    CACHE_TYPE_KEY = None  # type: Optional[str]

    def __init__(
        self,
        obj_type,
        obj_id,
        backend,
        storage,
        graph=None,
        objstorage=None,
        max_bundle_size=MAX_BUNDLE_SIZE,
    ):
        """Initialize the cooker.

        The type of the object represented by the id depends on the
        concrete class. Very likely, each type of bundle will have its
        own cooker class.

        Args:
            obj_type: type of the object to be cooked into a bundle (directory,
                      revision_flat or revision_gitfast; see
                      swh.vault.cooker.COOKER_TYPES).
            obj_id: id of the object to be cooked into a bundle.
            backend: the vault backend (swh.vault.backend.VaultBackend).
        """
        self.obj_type = obj_type
        self.obj_id = hashutil.hash_to_bytes(obj_id)
        self.backend = backend
        self.storage = storage
        self.objstorage = objstorage
        self.graph = graph
        self.max_bundle_size = max_bundle_size

    @abc.abstractmethod
    def check_exists(self):
        """Checks that the requested object exists and can be cooked.

        Override this in the cooker implementation.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def prepare_bundle(self):
        """Implementation of the cooker. Yields chunks of the bundle bytes.

        Override this with the cooker implementation.
        """
        raise NotImplementedError

    def cache_type_key(self) -> str:
        assert self.CACHE_TYPE_KEY
        return self.CACHE_TYPE_KEY

    def write(self, chunk):
        self.fileobj.write(chunk)

    def cook(self):
        """Cook the requested object into a bundle
        """
        self.backend.set_status(self.obj_type, self.obj_id, "pending")
        self.backend.set_progress(self.obj_type, self.obj_id, "Processing...")

        self.fileobj = BytesIOBundleSizeLimit(size_limit=self.max_bundle_size)
        try:
            try:
                self.prepare_bundle()
            except QueryCanceledError:
                raise PolicyError(
                    "Timeout reached while assembling the requested bundle"
                )
            bundle = self.fileobj.getvalue()
            # TODO: use proper content streaming instead of put_bundle()
            self.backend.put_bundle(self.cache_type_key(), self.obj_id, bundle)
        except PolicyError as e:
            self.backend.set_status(self.obj_type, self.obj_id, "failed")
            self.backend.set_progress(self.obj_type, self.obj_id, str(e))
        except Exception:
            self.backend.set_status(self.obj_type, self.obj_id, "failed")
            self.backend.set_progress(
                self.obj_type,
                self.obj_id,
                "Internal Server Error. This incident will be reported.",
            )
            logging.exception("Bundle cooking failed.")
        else:
            self.backend.set_status(self.obj_type, self.obj_id, "done")
            self.backend.set_progress(self.obj_type, self.obj_id, None)
        finally:
            self.backend.send_notif(self.obj_type, self.obj_id)



# Copyright (C) 2016  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import tarfile
import tempfile

from swh.model import hashutil
from swh.vault.cookers.base import BaseVaultCooker
from swh.vault.to_disk import DirectoryBuilder


class DirectoryCooker(BaseVaultCooker):
    """Cooker to create a directory bundle """

    CACHE_TYPE_KEY = "directory"

    def check_exists(self):
        return not list(self.storage.directory_missing([self.obj_id]))

    def prepare_bundle(self):
        with tempfile.TemporaryDirectory(prefix="tmp-vault-directory-") as td:
            directory_builder = DirectoryBuilder(self.storage, td.encode(), self.obj_id)
            directory_builder.build()
            with tarfile.open(fileobj=self.fileobj, mode="w:gz") as tar:
                tar.add(td, arcname=hashutil.hash_to_hex(self.obj_id))



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.model.model import Content, SkippedContent
from swh.vault.to_disk import get_filtered_files_content


def test_get_filtered_files_content(swh_storage):
    content = Content.from_data(b"foo bar")
    skipped_content = SkippedContent(
        sha1=None,
        sha1_git=b"c" * 20,
        sha256=None,
        blake2s256=None,
        length=42,
        status="absent",
        reason="for some reason",
    )
    swh_storage.content_add([content])
    swh_storage.skipped_content_add([skipped_content])

    files_data = [
        {
            "status": "visible",
            "sha1": content.sha1,
            "sha1_git": content.sha1_git,
            "target": content.sha1_git,
        },
        {"status": "absent", "target": skipped_content.sha1_git,},
    ]

    res = list(get_filtered_files_content(swh_storage, files_data))

    assert res == [
        {
            "content": content.data,
            "status": "visible",
            "sha1": content.sha1,
            "sha1_git": content.sha1_git,
            "target": content.sha1_git,
        },
        {
            "content": (
                b"This content has not been retrieved in the "
                b"Software Heritage archive due to its size."
            ),
            "status": "absent",
            "target": skipped_content.sha1_git,
        },
    ]


def test_get_filtered_files_content__unknown_status(swh_storage):
    content = Content.from_data(b"foo bar")
    swh_storage.content_add([content])

    files_data = [
        {
            "status": "visible",
            "sha1": content.sha1,
            "sha1_git": content.sha1_git,
            "target": content.sha1_git,
        },
        {"status": None, "target": b"c" * 20,},
    ]

    with pytest.raises(AssertionError, match="unexpected status None"):
        list(get_filtered_files_content(swh_storage, files_data))
